/* Copyright (c) 2015 Alinson Xavier
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _XOPEN_SOURCE 700

#include <math.h>
#include <stdio.h>
#include "double.h"
#include "util.h"
#include "lp.h"
#include "lifting.h"

int LFREE_2D_init(struct LFreeSet2D *set,
                  int n_vertices,
                  int n_lattice_points,
                  int n_halfspaces)
{
    int rval = 0;

    set->vertices = (double *) malloc(2 * n_vertices * sizeof(double));
    set->lattice_points = (double *) malloc( 2 * n_lattice_points * sizeof(double));
    set->halfspaces = (double *) malloc(2 * n_halfspaces * sizeof(double));

    abort_if(!set->vertices, "could not allocate set->vertices");
    abort_if(!set->lattice_points, "could not allocate set->lattice_points");
    abort_if(!set->halfspaces, "could not allocate set->halfspaces");

CLEANUP:
    return rval;
}

void LFREE_2D_free(struct LFreeSet2D *set)
{
    if(set->vertices) free(set->vertices);
    if(set->lattice_points) free(set->lattice_points);
    if(set->halfspaces) free(set->halfspaces);
}

/**
 * Verifies if the set is well formed.
 */
int LFREE_2D_verify(struct LFreeSet2D *set)
{
    int rval = 0;

    abort_if(set->n_halfspaces == 0, "Halfspaces not found");

    for(int i = 0; i < set->n_lattice_points; i++)
    {
        double *t = &set->lattice_points[2 * i];
        double r[2] = {t[0] - set->f[0],
                       t[1] - set->f[1]};

        double value;

        rval = LIFTING_2D_psi(set->n_halfspaces, set->halfspaces, r, &value);
        abort_if(rval, "LIFTING_2D_psi failed");

        double delta = fabs(value - 1);
        if(delta > 0.0001)
        {
            log_debug("Lattice point (%.2lf, %.2lf) is "
                "not on the boundary (delta=%.6lf)", t[0], t[1], delta);
            rval = 1;
            goto CLEANUP;
        }
    }

CLEANUP:
    return rval;
}

/**
 * Computes the bounding box around the given set.
 *
 * @param[in]  set  the set whose bounding box should be computed
 * @param[out] lb   two-dimensional vector representing lower-left corner
 *                  of the bounding box
 * @param[out] ub   two-dimensional vector representing upper-right corner
 *                  of the bounding box
 */
int LFREE_2D_get_bounding_box(const struct LFreeSet2D *set,
                              int *lb,
                              int *ub)
{
    int rval = 0;

    ub[0] = ub[1] = INT_MIN;
    lb[0] = lb[1] = INT_MAX;

    for (int i = 0; i < set->n_vertices; i++)
    {
        double *vertex = &set->vertices[2 * i];

        ub[0] = (int) fmax(ub[0], ceil(vertex[0]));
        ub[1] = (int) fmax(ub[1], ceil(vertex[1]));

        lb[0] = (int) fmin(lb[0], floor(vertex[0]));
        lb[1] = (int) fmin(lb[1], floor(vertex[1]));
    }

CLEANUP:
    return rval;
}

int LFREE_2D_print_set(const struct LFreeSet2D *set)
{
    int rval = 0;

    log_debug("  f=%12.6lf %12.6lf\n", set->f[0], set->f[1]);

    for (int i = 0; i < set->n_vertices; i++)
    {
        double *vertex = &set->vertices[2 * i];
        log_debug("  v%-3d=%12.6lf %12.6lf\n", i, vertex[0], vertex[1]);
    }

    for (int i = 0; i < set->n_lattice_points; i++)
    {
        double *lattice_point = &set->lattice_points[2 * i];
        log_debug("  t%-3d=%12.6lf %12.6lf\n", i, lattice_point[0], lattice_point[1]);
    }

    for(int i = 0; i < set->n_halfspaces; i++)
    {
        double *halfspace = &set->halfspaces[2 * i];
        log_debug("  h%-3d=%12.6lf %12.6lf\n", i, halfspace[0], halfspace[1]);
    }

CLEANUP:
    return rval;
}

/**
 * Computes the halfspace representation for a given set.
 *
 * To construct a LFreeSet2D, it is only necessary to specify
 * the vertices and the interior point. This function computes
 * the halfspace representation. It assumes that the vertices are
 * sorted in either clockwise or counter-clockwise order.
 * The set is modified in-place.
 *
 * @param set  the set whose halfspace representation should
 *             be computed
 *
 */
int LFREE_2D_compute_halfspaces(struct LFreeSet2D *set)
{
    int rval = 0;

    int k = 0;
    int expected_sgn = 0;
    double *f = set->f;

    for (int i0 = 0; i0 < set->n_vertices; i0++)
    {
        int i1 = (i0 + 1) % set->n_vertices;
        int i2 = (i0 + 2) % set->n_vertices;

        double *v0 = &set->vertices[2 * i0];
        double *v1 = &set->vertices[2 * i1];
        double *v2 = &set->vertices[2 * i2];

        double x0 = v0[0], y0 = v0[1];
        double x1 = v1[0], y1 = v1[1];
        double x2 = v2[0], y2 = v2[1];

        int sgn = DOUBLE_sgn(
                -x1 * y0 + x2 * y0 + x0 * y1 - x2 * y1 - x0 * y2 + x1 * y2);

        abort_if(sgn == 0, "vertices should not be aligned");
        if(expected_sgn == 0) expected_sgn = sgn;
        abort_if(expected_sgn != sgn, "wrong orientation of vertices");

        double *halfspace = &set->halfspaces[2 * (k++)];

        halfspace[0] = y0 - y1;
        halfspace[1] = x1 - x0;

        double rhs = y0 * x1 - x0 * y1;
        rhs -= (y0 - y1) * f[0];
        rhs -= (x1 - x0) * f[1];

        halfspace[0] /= rhs;
        halfspace[1] /= rhs;
    }

    set->n_halfspaces = k;

CLEANUP:
    return rval;
}

/**
 * Translates a given set by a certain amount.
 *
 * @param set  the set to be translated
 * @param dx   offset to be applied to the first component
 * @param dy   offset to be applied to the second component
 */
int LFREE_2D_translate_set(struct LFreeSet2D *set, double dx, double dy)
{
    set->f[0] += dx;
    set->f[1] += dy;

    for (int i = 0; i < set->n_vertices; i++)
    {
        double *vertex = &set->vertices[2 * i];
        vertex[0] += dx;
        vertex[1] += dy;
    }

    for (int i = 0; i < set->n_lattice_points; i++)
    {
        double *lattice_point = &set->lattice_points[2 * i];
        lattice_point[0] += dx;
        lattice_point[1] += dy;
    }

    return 0;
}

/**
 * Computes the inverse of the given matrix and left-multiplies it
 * by the given ray. The matrix is given by [[x1,y1],[x2,y2]]. The
 * ray is modified in-place.
 */
static void apply_inverse(double x1,
                          double y1,
                          double x2,
                          double y2,
                          double *ray)
{
    double a = ray[0];
    double b = ray[1];
    double det = x2 * y1 - x1 * y2;

    ray[0] = (b * x2 - a * y2) / det;
    ray[1] = (a * y1 - b * x1) / det;
}

/**
 * Applies a unimodular affine transformation to the set, making
 * lattice points 1 and 2 orthogonal. Returns the transformation
 * applied.
 *
 * @param[in] set the set to be transformed
 * @param[out] m two-by-two matrix representing the transformation
 *               applied
 */
static int make_t1_t2_orthogonal(struct LFreeSet2D *set, double *m)
{
    int rval = 0;

    double x1 = set->lattice_points[2];
    double y1 = set->lattice_points[3];

    double x2 = set->lattice_points[4];
    double y2 = set->lattice_points[5];

    double det = x2 * y1 - x1 * y2;
    abort_if(DOUBLE_iszero(det), "determinant should not be zero");

    apply_inverse(x1, y1, x2, y2, set->f);

    for (int i = 0; i < set->n_vertices; i++)
    {
        double *vertex = &set->vertices[2 * i];
        apply_inverse(x1, y1, x2, y2, vertex);
    }

    for (int i = 0; i < set->n_lattice_points; i++)
    {
        double *lattice_point = &set->lattice_points[2 * i];
        apply_inverse(x1, y1, x2, y2, lattice_point);
    }

    apply_inverse(x1, y1, x2, y2, &m[0]);
    apply_inverse(x1, y1, x2, y2, &m[2]);

CLEANUP:
    return rval;
}

static int apply_m_to_vect(double *v, const double *m)
{
    double v0 = v[0], v1 = v[1];
    v[0] = m[0] * v0 + m[1] * v1;
    v[1] = m[2] * v0 + m[3] * v1;

    return 0;
}

/**
 * Applies a unimodular affine transformation to the set, making
 * the four lattice points the unit square.
 *
 * Must be called on quadrilaterals only, and assumes that the three
 * first lattice points already correspond to (0,0), (0,1) and (1,0).
 * The set is modified in place. The transformation is also applied
 * to a given matrix pre_m.
 */
static int make_square(struct LFreeSet2D *set, double *pre_m)
{
    int rval = 0;
    double m[4];

    abort_if(set->n_vertices != 4, "make_square requires quadrilaterals");

    double *t = &set->lattice_points[2 * 3];
    if(t[0] == 1 && t[1] == 1)
    {
        m[0] = 1; m[1] = 0;
        m[2] = 0; m[3] = 1;
    }
    else if(t[0] == 1 && t[1] == -1)
    {
        m[0] = 1; m[1] = 0;
        m[2] = 1; m[3] = 1;
    }
    else if(t[0] == -1 && t[1] == 1)
    {
        m[0] = 1; m[1] = 1;
        m[2] = 0; m[3] = 1;
    }
    else
    {
        abort_if(1, "invalid state");
    }

    rval = LFREE_2D_transform_set(set, m);
    abort_if(rval, "LFREE_2D_transform_set failed");

    apply_m_to_vect(&pre_m[0], m);
    apply_m_to_vect(&pre_m[2], m);

CLEANUP:
    return rval;
}


int find_thin_direction(const struct LFreeSet2D *set,
                        double *thin_direction,
                        double *width)
{
    int rval = 0;

    int n_candidates = 3;
    double candidates[] = {
            0,  1,
            1,  0,
            1,  1
    };

    double best_width = INFINITY;
    double *best_candidate = 0;

    for (int i = 0; i < n_candidates; i++)
    {
        double *d = &candidates[2 * i];
        double max_val = -INFINITY;
        double min_val = INFINITY;

        for (int j = 0; j < set->n_vertices; j++)
        {
            double *vertex = &set->vertices[2 * j];
            double val = vertex[0] * d[0] + vertex[1] * d[1];

            max_val = fmax(max_val, val);
            min_val = fmin(min_val, val);
        }

        double delta = fabs(max_val - min_val);

        log_verbose("d=%.3lf %.3lf width=%.3lf\n", d[0], d[1], delta);

        if(delta < best_width)
        {
            best_width = delta;
            best_candidate = d;
        }
    }

    memcpy(thin_direction, best_candidate, 2 * sizeof(double));
    *width = best_width;

CLEANUP:
    return rval;
}

int LFREE_2D_transform_set(struct LFreeSet2D *set, const double *m)
{
    int rval = 0;

    rval = apply_m_to_vect(set->f, m);
    abort_if(rval, "apply_m_to_vect failed");

    for (int i = 0; i < set->n_vertices; i++)
    {
        double *vertex = &set->vertices[2 * i];

        rval = apply_m_to_vect(vertex, m);
        abort_if(rval, "apply_m_to_vect failed");
    }

    for (int i = 0; i < set->n_lattice_points; i++)
    {
        double *lattice_point = &set->lattice_points[2 * i];

        rval = apply_m_to_vect(lattice_point, m);
        abort_if(rval, "apply_m_to_vect failed");
    }

CLEANUP:
    return rval;
}

int LFREE_2D_preprocess(struct LFreeSet2D *set, double *pre_m, double *center)
{
    int rval = 0;
    double direction[2], m[4], width;
    double translate[2];

    pre_m[0] = 1; pre_m[1] = 0;
    pre_m[2] = 0; pre_m[3] = 1;

    rval = LFREE_2D_translate_set(set, -set->lattice_points[0],
            -set->lattice_points[1]);
    abort_if(rval, "LFREE_2D_translate set failed");

    rval = make_t1_t2_orthogonal(set, pre_m);
    abort_if(rval, "make_t1_t2_orthogonal failed");

    if(set->n_vertices == 4)
    {
        rval = make_square(set, pre_m);
        abort_if(rval, "make_square failed");
    }

    rval = find_thin_direction(set, direction, &width);
    abort_if(rval, "find_thin_direction failed");

    log_debug("d=%.2lf %.2lf\n", direction[0], direction[1]);

    if(direction[0] == 0 && direction[1] == 1)
    {
        m[0] = 1; m[1] = 0;
        m[2] = 0; m[3] = 1;
        translate[0] = translate[1] = 0;
    }
    else if(direction[0] == 1 && direction[1] == 0)
    {
        m[0] = 0; m[1] = 1;
        m[2] = 1; m[3] = 0;
        translate[0] = translate[1] = 0;
    }
    else if(direction[0] == 1 && direction[1] == 1)
    {
        m[0] =  1; m[1] =  0;
        m[2] =  -1; m[3] =  -1;
        translate[0] = 0;
        translate[1] = 1;
    }
    else
    {
        abort_if(1, "invalid direction");
    }

    apply_m_to_vect(&pre_m[0], m);
    apply_m_to_vect(&pre_m[2], m);

    rval = LFREE_2D_transform_set(set, m);
    abort_if(rval, "LFREE_2D_transform_set failed");

    rval = LFREE_2D_translate_set(set, translate[0], translate[1]);
    abort_if(rval, "LFREE_2D_translate_set failed");

    double max_v1 = -INFINITY;
    double min_v1 = INFINITY;

    for(int i = 0; i < set->n_vertices; i++)
    {
        double *v = &set->vertices[2 * i];
        max_v1 = fmax(max_v1, v[1]);
        min_v1 = fmin(min_v1, v[1]);
    }

    *center = (max_v1 + min_v1) / 2;
    width = max_v1 - min_v1;

    log_debug("width = %.2lf\n", width);
    abort_iff(width >= 3, "lattice width too large: %.2lf", width);

CLEANUP:
    return rval;
}

int LFREE_2D_read_next(FILE *file, struct LFreeSet2D *set)
{
    int rval = 0;
    ssize_t count;

    double *f = set->f;
    count = fscanf(file, "%lf %lf ", &f[0], &f[1]);
    abort_if(count != 2, "could not read f");

    count = fscanf(file, "%d ", &set->n_vertices);
    abort_if(count != 1, "could not read n_vertices");
    abort_if(set->n_vertices != 3 && set->n_vertices != 4,
            "only triangles or quadrilaterals supported");

    for(int i = 0; i < set->n_vertices; i++)
    {
        double *vertex = &set->vertices[2 * i];
        count = fscanf(file, "%lf %lf ", &vertex[0], &vertex[1]);
        abort_iff(count != 2, "could not read vertex %d", i+1);
    }

    count = fscanf(file, "%d ", &set->n_lattice_points);
    abort_if(count != 1, "could not read n_lattice_points");
    abort_if(set->n_lattice_points != set->n_vertices, "number of vertices "
            "should be the same as number of lattice points");

    for(int i = 0; i < set->n_lattice_points; i++)
    {
        double *lattice_point = &set->lattice_points[2 * i];
        count = fscanf(file, "%lf %lf ", &lattice_point[0], &lattice_point[1]);
        abort_iff(count != 2, "could not read lattice_point %d", i+1);
    }

CLEANUP:
    return rval;
}

static int create_continuous_lp(struct LP *lp,
                                int dim,
                                int n_halfspaces,
                                int n_fixed_vars,
                                double *halfspaces,
                                double *fixed_vars)
{
    int rval = 0;

    int rmatbeg = 0;
    int* rmatind = 0;
    double *rmatval = 0;

    rmatind = (int *) malloc((dim + 1) * sizeof(int));
    rmatval = (double *) malloc((dim + 1) * sizeof(double));
    abort_if(!rmatind, "could not allocate rmatind");
    abort_if(!rmatval, "could not allocate rmatval");

    rval = LP_create(lp, "lifting_continuous");
    abort_if(rval, "LP_create failed");

    // create epsilon variable
    rval = LP_new_col(lp, 1.0, -MILP_INFINITY, MILP_INFINITY, 'C');
    abort_if(rval, "LP_new_col failed");

    // create alpha variables
    for (int i = 0; i < dim - n_fixed_vars; i++)
    {
        rval = LP_new_col(lp, 0.0, -MILP_INFINITY, MILP_INFINITY, 'C');
        abort_if(rval, "LP_new_col failed");
    }

    // add constraints: for i=1,...,m:
    //     e - \sum_{j=1}^m a^i_j alpha_j \geq
    //     \sum_{j=n-l+1}^n a^i_j \bar alpha_j
    for (int i = 0; i < n_halfspaces; i++)
    {
        char sense = 'G';
        double rhs = 0.0;
        double *ai = &halfspaces[i * dim];

        for(int j = 0; j < n_fixed_vars; j++)
            rhs += ai[dim - n_fixed_vars + j] * fixed_vars[j];

        int nz = 0;

        rmatind[nz] = 0;
        rmatval[nz] = 1.0;
        nz++;

        for (int j = 0; j < dim - n_fixed_vars; j++)
        {
            rmatind[nz] = j + 1;
            rmatval[nz] = - ai[j];
            nz++;
        }

        rval = LP_add_rows(lp, 1, nz, &rhs, &sense, &rmatbeg, rmatind, rmatval);
        abort_if(rval, "LP_add_rows failed");
    }

    //rval = LP_write(lp, "continuous.lp");
    //abort_if(rval, "LP_write failed");

CLEANUP:
    if (rmatind) free(rmatind);
    if (rmatval) free(rmatval);
    return rval;
}

int LIFTING_2D_optimize_continuous(int n_halfspaces,
                                   const double *halfspaces,
                                   double alpha2,
                                   double *alpha1,
                                   double *value)
{
    int rval = 0;

    *value = -INFINITY;

    for(int r = 0; r < n_halfspaces; r++)
    {
        for(int s = r+1; s < n_halfspaces; s++)
        {
            const double *ar = &halfspaces[r * 2];
            const double *as = &halfspaces[s * 2];

            double xr = ar[0];
            double xs = as[0];
            if(DOUBLE_eq(xr, xs)) continue;

            double br = - xs  / (xr - xs);
            double bs = xr / (xr - xs);
            if(!DOUBLE_geq(br, 0)) continue;
            if(!DOUBLE_geq(bs, 0)) continue;

            double yr = ar[1] * alpha2;
            double ys = as[1] * alpha2;

            double obj = yr * br + ys * bs;

            if(DOUBLE_geq(obj, *value))
            {
                *value = obj;

                if(DOUBLE_neq(xr, 0))
                    *alpha1 = - (yr - obj) / xr;
                else
                    *alpha1 = - (ys - obj) / xs;
            }
        }
    }

CLEANUP:
    return rval;
}

int LIFTING_ND_optimize_continuous(int dim,
                                   int n_halfspaces,
                                   int n_fixed_vars,
                                   double *halfspaces,
                                   double *fixed_vars,
                                   double *opt)
{
    int rval = 0;

    struct LP lp;

    rval = LP_open(&lp);
    abort_if(rval, "LP_open failed");

    rval = create_continuous_lp(&lp, dim, n_halfspaces, n_fixed_vars,
            halfspaces, fixed_vars);
    abort_if(rval, "create_continuous_lp failed");

    int infeasible;
    rval = LP_optimize(&lp, &infeasible);
    abort_if(rval, "LP_optimize failed");

    abort_if(infeasible, "LP should alway be feasible");

    rval = LP_get_obj_val(&lp, opt);
    abort_if(rval, "LP_get_obj_val failed");

CLEANUP:
    return rval;
}

int LIFTING_2D_psi(int n_halfspaces,
                   const double *halfspaces,
                   const double *ray,
                   double *value)
{
    int rval = 0;

    *value = -INFINITY;

    for(int i = 0; i < n_halfspaces; i++)
    {
        const double *h = &halfspaces[2 * i];
        double v = ray[0] * h[0] + ray[1] * h[1];
        *value = DOUBLE_max(v, *value);
    }

CLEANUP:
    return rval;
}

int LIFTING_2D_lift_fixed(int n_halfspaces,
                          const double *halfspaces,
                          const double *ray,
                          double k1,
                          double *opt)
{
    int rval = 0;

    double k0;
    double value;

    rval = LIFTING_2D_optimize_continuous(n_halfspaces, halfspaces,
            ray[1] + k1, &k0, &value);
    abort_if(rval, "LIFTING_2D_optimize_continuous failed");

    double delta = k0 - ray[0];

    double r_ceil[2] = { ray[0] + ceil(delta), ray[1] + k1 };
    double r_floor[2] = { ray[0] + floor(delta), ray[1] + k1 };

    log_debug("    delta=%.6lf value=%.6lf\n", delta, value);
    log_debug("    r_ceil=%12.6lf %12.6lf\n", r_ceil[0], r_ceil[1]);
    log_debug("    r_floor=%12.6lf %12.6lf\n", r_floor[0], r_floor[1]);

    double value_ceil, value_floor;

    rval = LIFTING_2D_psi(n_halfspaces, halfspaces, r_ceil, &value_ceil);
    abort_if(rval, "LIFTING_2D_psi failed");

    rval = LIFTING_2D_psi(n_halfspaces, halfspaces, r_floor, &value_floor);
    abort_if(rval, "LIFTING_2D_psi failed");

    *opt = min(value_ceil, value_floor);

CLEANUP:
    return rval;
}

int LIFTING_2D_naive(int n_halfspaces,
                     const double *halfspaces,
                     const double *ray,
                     const int *lb,
                     const int *ub,
                     double *value)
{
    int rval = 0;

    *value = INFINITY;
    int best_k0 = 0;
    int best_k1 = 0;

    int margin = 10;

    for(int k0 = lb[0] - margin; k0 <= ub[0] + margin; k0++)
    {
        for(int k1 = lb[1] - margin; k1 <= ub[1] + margin; k1++)
        {
            double q[2] = { ray[0] + k0, ray[1] + k1 };
            double value_q;

            rval = LIFTING_2D_psi(n_halfspaces, halfspaces, q, &value_q);
            abort_if(rval, "LIFTING_2D_ps failed");

            if(value_q < *value)
            {
                best_k0 = k0;
                best_k1 = k1;
                *value = value_q;
                log_debug("  k=%6d %6d value=%12.6lf\n", k0, k1, *value);
            }
        }
    }

//    log_debug("      best_k=(%d %d) value=%.6lf\n", best_k0, best_k1, *value);

CLEANUP:
    return rval;
}

int LIFTING_2D_bound(int n_halfspaces,
                     const double *halfspaces,
                     const double *ray,
                     double *value)
{
    int rval = 0;

    double eta_star, eta_plus, eta_minus;
    double xi_plus, xi_minus;
    double m_plus, m_minus;
    double ignored;
    int k1 = 1;

    rval = LIFTING_2D_lift_fixed(n_halfspaces, halfspaces, ray, 0, &eta_star);
    abort_if(rval, "LIFTING_2D_lift_fixed failed");

    rval = LIFTING_2D_optimize_continuous(n_halfspaces, halfspaces, 1, &ignored,
            &xi_plus);
    abort_if(rval, "LIFTING_2D_optimize_continuous failed");

    rval = LIFTING_2D_optimize_continuous(n_halfspaces, halfspaces, -1,
            &ignored, &xi_minus);
    abort_if(rval, "LIFTING_2D_optimize_continuous failed");

    m_plus = m_minus = INFINITY;

    log_debug("Level 0:\n");
    log_debug("  eta star  = %.6lf\n", eta_star);
    log_debug("  next slack plus  = %.6lf\n", k1 + ray[1] - m_plus);
    log_debug("  next slack minus = %.6lf\n", k1 - ray[1] - m_minus);
    log_debug("  xi plus = %.6lf\n", xi_plus);
    log_debug("  xi minus = %.6lf\n", xi_minus);

    while((k1 + ray[1] <= m_plus) || (k1 - ray[1] <= m_minus))
    {
        log_debug("Level %d:\n", k1);

        double h_plus, h_minus;

        rval = LIFTING_2D_optimize_continuous(n_halfspaces, halfspaces, ray[1]
                + k1, &ignored, &h_plus);
        abort_if(rval, "LIFTING_2D_optimize_continuous failed");

        rval = LIFTING_2D_optimize_continuous(n_halfspaces, halfspaces, ray[1]
                - k1, &ignored, &h_minus);
        abort_if(rval, "LIFTING_2D_optimize_continuous failed");

        rval = LIFTING_2D_lift_fixed(n_halfspaces, halfspaces, ray, k1,
                &eta_plus);
        abort_if(rval, "LIFTING_2D_lift_fixed failed");

        rval = LIFTING_2D_lift_fixed(n_halfspaces, halfspaces, ray, -k1,
                &eta_minus);
        abort_if(rval, "LIFTING_2D_lift_fixed failed");

        eta_star = fmin(eta_star, fmin(eta_plus, eta_minus));

        m_plus = ceil(eta_star / xi_plus);
        m_minus = ceil(eta_star / xi_minus);

        log_debug("  eta plus  = %.6lf\n", eta_plus);
        log_debug("  eta minus = %.6lf\n", eta_minus);
        log_debug("  eta star  = %.6lf\n", eta_star);
        log_debug("    h minus = %.6lf\n", h_minus);
        log_debug("     h plus = %.6lf\n", h_plus);

        k1++;

        log_debug("  next slack plus  = %.6lf\n", k1 + ray[1] - m_plus);
        log_debug("  next slack minus = %.6lf\n", k1 - ray[1] - m_minus);
    }

    log_debug("Done\n");

    *value = eta_star;

CLEANUP:
    return rval;
}
